/* Copyright 2000 John Fremlin
   Distributed under GPL, see file COPYING
*/

/* This code is NOT THREAD SAFE (it uses strerror) */

#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <termio.h>
#include <signal.h>
#include "config.h"
#include "system.h"

#ifndef USE_TERMINAL_PIDS
#define USE_TERMINAL_PIDS 1
#endif

inline
void
pids_dump(void)
{
  pid_t pid = getpid();

  fprintf(stderr,gt("%s: (from %s %s) dumping pid information\n"),
	  get_program_invocation_name(),PACKAGE,VERSION);
  
  if(!isatty(0))
    {
      fprintf(stderr,gt("%s: warning: fd 0 is not a tty\n"),
	      get_program_invocation_name());
    }

#if USE_TERMINAL_PIDS
  fprintf(stderr,gt("%s: %d is terminal controlling process group\n"),
	  get_program_invocation_name(),tcgetpgrp(0));
  fprintf(stderr,gt("%s: %d is terminal controlling session\n"),
	  get_program_invocation_name(),tcgetsid(0));
#endif

  fprintf(stderr,gt("%s: %d is process group\n"),
	  get_program_invocation_name(),getpgrp());
  fprintf(stderr,gt("%s: %d is session\n"),
	  get_program_invocation_name(),getsid(pid));
  fprintf(stderr,gt("%s: %d is process\n"),
	  get_program_invocation_name(),pid);

}
#if 0
void own_tty(int fd) 
{
  pid_t pid=getpid();
  
  if(ioctl(fd,TIOCSCTTY,pid))
    {
      fprintf(stderr,gt("%s: warning: can't become tty session controller: %s\n"),
	      get_program_invocation_name(),strerror(errno));
    }
  
  /* can't change pgid of terminal unless you're in the right session
     group, so this usually fails unless TIOSCSTTY succeeded first */
  if(tcsetpgrp(fd, pid))
    {
      fprintf(stderr,gt("%s: warning: can't tcsetpgrp: %s\n"),
	      get_program_invocation_name(),strerror(errno));
    }
}

void
reset_tty(int fd)
{
#ifdef HAVE_VHANGUP
  struct sigaction old, off;
  sigemptyset (&off.sa_mask);
  off.sa_flags = 0;
  off.sa_handler = SIG_IGN;
  if(sigaction (SIGHUP, &off,&old)<0){
    perror("sigaction");
    return;
  }

  if(vhangup()<0) /* wastes all fds pointing to tty */
    {
      close(fd);
      perror("vhangup");
    }
  
  if(sigaction(SIGHUP,&old,0)<0){
    perror("restoring sigaction");
    return;
  }
#else
  close(fd);
#endif
}
#endif

static
int
open_term(const char*device)
{

  /* Test we can open the tty. Then we close all fds and open again,
     so that it becomes our controlling terminal even if we already
     had one */
  int fd = open(device,O_RDWR|O_NONBLOCK,0);
  int i;
  
  if(fd==-1)
    {
      fprintf(stderr,gt("%s: problem opening device %s: %s\n"),
	      get_program_invocation_name(),device,strerror(errno));
      return 1;
    }

  if(!isatty(fd))
    {
      fprintf(stderr,gt("%s: warning: %s is not a tty\n"),
	      get_program_invocation_name(),device);
    }

  for(i=0;i<=fd;i++)
    close(i);

  fd=open(device,O_RDWR,0);
  if(fd==-1)abort();
  
  for(i=0;i<3;i++)
    if(fd!=i)if(dup(fd)!=i)abort();

  if(fd>2)close(fd);
  
  return 0;
}

static
int
pids_admin(void)
{
  pid_t pid=getpid();
  pid_t session=getsid(pid);
  pid_t group=getpgid(pid);
  
  if(pid==session)/*someone (init?) has kindly given us our own session*/
    {
      if(group!=pid)
	setpgid(pid,pid);/*but not our own process group*/
    }
  else
    {
      /* We can't setsid successfully if we already are leaders of a
	 process group (process id == process group id), so we fork for a
	 new pid (which will end up in the current process group) */

      if(group==pid)
	{
	  pid_t child=fork();
	  if(child)
	    {
	      /* in parent: to wait or not to wait? init would go
		 respawn mad if we don't, but init is not like a
		 shell; it gives us our own session. If we do wait we
		 waste a process entry. */
	      exit(0);
	    }
	  
	  pid=getpid();
	  /* in child */
	  if(setsid()==-1)
	    /* note that setsid gives us our own process group as well */
	    {
	      fprintf(stderr,gt("%s: warning: can't setsid: %s\n"),
		      get_program_invocation_name(),strerror(errno));
	      pids_dump();
	    }
	}
    }
  
#if 0  
  /* On vhangup called from login, we get killed (login blocks the
     signal as below, so it's ok if we've exec'd login) */
  signal(SIGHUP,SIG_IGN);
#endif
  
  return 0;
}  

int
main(int argc,char*argv[],char*env[])
{
  gettext_init();
  
  if(argc<2)
    {
      fprintf(stderr,gt("%s: too few arguments\n"),get_program_invocation_name());
      printf(gt("%s from %s %s\t\t(c) 2000 John Fremlin\n"), get_program_invocation_short_name(), PACKAGE, VERSION);
      printf(gt("syntax: %s [terminal device] [optional program to run]\n\n"),get_program_invocation_short_name());
      printf(gt("If no other program is specified, %s is run.\n"),DEFAULT_AGENT);
      return 1;
    }
  
  if(pids_admin())
    return 2;

  if(open_term(argv[1]))
    return 3;
  
  if(argv[2])
    {
      if(execvp(argv[2],argv+2))
	{
	  fprintf(stderr,gt("%s: problem execvping %s: %s\n"),get_program_invocation_name(),argv[2],strerror(errno));
	  return 4;
	}
    }
  else
    {
      if(execl(DEFAULT_AGENT,DEFAULT_AGENT,0))
	{
	   fprintf(stderr,gt("%s: problem execling login agent %s: %s\n"),get_program_invocation_name(),DEFAULT_AGENT,strerror(errno));
	   return 5;
	}
    }
  
  return 0;
}

